/*	##############################################################################################
	Advanced RTI System, ARTÌS			http://pads.cs.unibo.it

	Description:	
		Airports example. 3 LP that model 3 interconnected airports. 
		In this model the synchronization mechanism is: Timestepped

	Output:
		the mode output can be found in the <LP_ID>.out file
		the middleware output is put on screen

	Changelog:
		14/10/09
			A few more comments

	Authors:
		First version by Michele Bracuto <bracuto@cs.unibo.it>
		Modified by Gabriele D'Angelo <gda@cs.unibo.it>

	############################################################################################## */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>

//	ARTÌS - Timestepped header file
#include <ts.h>			

//	Model-related constants
#define FLIGHT_T		4	// time required to flight from an airport to another
#define LOAD_T			2	// time required to load/unload passengers
#define LAND_T			1	// landing time, required to free the runway
#define WAIT_T			1	// time a plane has to wait before re-requesting landing
#define PLANES_NUMBER		2	// airplanes initially in the hangar

//	Timestepped error type
extern	char	ts_error[MAX_ERROR_STRING];

/*------------------GLOBAL STATE VARIABLES------------------------*/

//	Time management variables
short	end_reached	= 0;
double	end_clock	= 2000.0;	// end of the simulated (simulated time)
static	double	clock	= 0.0;		// current simulated time
// 	Seed used for the random generator
double	SEED;


/* ------------------SIMULATOR STATE VARIABLES-----------------*/

int	runway_isfree;					// 1 if the runway is free, 0 in other cases
int	on_the_ground;					// number of airplanes on the ground
int	in_the_air;					// number of airplanes on air
int	NEIGHBOUR1, NEIGHBOUR2, LP;			// identifiers
char	*cnames[3]={"MILAN", "BOLOGNA", "ROME"};	// aiport names


/*-----------MESSAGE STRUCTURE AND CONSTRUCTOR-----------*/

//	Structure of messages
typedef struct _sub_msg_header {
	char   event_type[4];	// type of message (model related)
	char   from[10];	// source of the message (airport name)
	char   id[12];		// flight identifier
} SubMsgHeader;


/*-------------------MESSAGE CREATION-----------------------------*/
char 	*create_submsg (char type[], char from[], char *id) {
	SubMsgHeader  * tmsg;

	tmsg = (SubMsgHeader *)malloc (sizeof(SubMsgHeader));
	memset (tmsg, 0, sizeof(SubMsgHeader));
	strcpy (tmsg->event_type, type);
	strcpy (tmsg->from, from);
	strcpy (tmsg->id, id);

	return (char *)tmsg;
}


/*-------------------RANDOM NUMBERS GENERATOR----------------------*/

//	Seeds used by the generator
double SEEDS[20] = {
        606112567, 2005261598, 975679233, 381883932, 1949009736,
        457621696, 31040274, 2084138466, 63789657, 521636194,
        670018843, 1263912019, 2072256297, 1025373591, 1693757298,
        1115383532, 44685244, 1568381214, 2014569565, 1052870108
};

//	Random numbers generator
static double my_rand ()
{
	static long m = 2147483647;
	static long a = 48271;
	static long q = 44488;
	static long r = 3399;
	long t, lo, hi;

	hi 	= SEED / q;
	lo 	= SEED - q * hi;
	t 	= a * lo - r * hi;

	if (t > 0)
		SEED = t;
	else
		SEED = t + m;


	return ( (double) SEED / (double) m );
}


//	Returns 0 or 1
static int	rnd_bool (void)
{
	return ( my_rand() <= 0.5 ? 0 : 1 );
}


/*----------------------------EVENT HANDLERS------------------------*/

//	Arrival (to the local LP) of an airplane
void	arrival_event_handler (int src, char *from, char *id)
{
	//	Check the runway
	if (runway_isfree){
		//	It is free
		in_the_air++;		// one more airplane is waiting for landing
		runway_isfree = 0;	// lock the runway

		//	Schedule the landing of the airplane
		TS_Schedule(clock+LAND_T, (void *)create_submsg ("LAN", from, id), sizeof(SubMsgHeader));
		printf ("[%s]:   %.1f  Waiting for landing, flight %s from %s\n", cnames[LP], clock, id, from);
		fflush(stdout);
	}
	else {
		//	If the source of the plane is not the local airport, then 
		//	one more airplane is waiting for landing
		if(src != LP)	in_the_air++;

		//	Schedule a new arrival event in future	
		TS_Schedule(clock+WAIT_T, (void *)create_submsg ("ARR", from, id), sizeof(SubMsgHeader));
		printf ("[%s]:   Runway busy, the flight %s from %s is waiting\n", cnames[LP], id, from);
		fflush(stdout);
	}	
}

//	Landing (to the local LP) of an airplane
void	landed_event_handler (char *from, char *id)
{
	//	One less on air, one more on the ground
	in_the_air--;
	on_the_ground++;

	//	Schedule the next departure (of the plane)
	TS_Schedule (clock + LOAD_T, (void *)create_submsg ("DEP", "", id), sizeof(SubMsgHeader));
	
	printf ("[%s]:     %.1f  Landed flight %s from %s\n", cnames[LP], clock, id, from);
	fflush(stdout);

	//	Now the runway is free
	runway_isfree = 1;
}

//	Departure (from the local LP) of an airplane
void	departure_event_handler (char *id)
{
	int dest;	// destination airport (LP)
	int ret;	// return value
	
	//	One less on the ground
	on_the_ground--;

	//	Choose a random destination from the 2 available
	dest = rnd_bool() == 0 ? NEIGHBOUR1 : NEIGHBOUR2;	
	 
	//	Send to the destination airport (LP) the arrival event
	ret = TS_Send(dest, clock + FLIGHT_T, (void *)create_submsg ("ARR", cnames[LP], id), sizeof(SubMsgHeader) );
	
	//	Check the sending operation
	if(ret >= 0) {
		//	If OK then trace out the departure
		printf ("[%s]: %.1f Departed flight %s with destination %s\n",  cnames[LP], clock, /*clock + FLIGHT_T,*/ id, cnames[dest]);
		fflush(stdout);
	}
	else {
		//	Else print an error message
		printf(ts_error);
		fflush(stdout);
	}
}


/*----------------------------MAIN FUNCTION----------------------------*/

int main(int argc, char* argv[])
{
	SubMsgHeader  	*submsg;	// event message

	char	msg[1024];		// message buffer
	char 	id[20];			// airplane identification string
	int	from;			// source LP (of a message)
	long 	size;			// message size
	double	Ts;			// time variable
	int	total_flies = 0;	// total number of departed flies (from this airport)
	int 	i;			// temporary variable
	char	out[100]; 		// Output buffer
	
	
	//	Setup the TIMESTEPPED synchronization algorithm
	LP = TS_Init(cnames[atoi(argv[1])], argv[2], 5000);

	// Redirection of standard output and error
	sprintf(out, "%d.out", LP);
	fclose(stdout); stdout = fopen(out, "w"); fflush(stdout);	
	
	//	Initialize the airports identifiers
	NEIGHBOUR1	= (LP + 1) % 3;
	NEIGHBOUR2	= (LP + 2) % 3;
	printf(" %s <- [%s] -> %s\n", cnames[NEIGHBOUR1], cnames[LP], cnames[NEIGHBOUR2] );

	//	Random numbers generator, seed initialization
	SEED		= SEEDS[LP];

	//	Model variables initialization
	runway_isfree 	= 1;	// the runway is free
   	on_the_ground 	= 2;	// 2 planes (in each airport) are on the ground
   	in_the_air 	= 0;	// no airplanes are on the air

	//	Boostrapping the simulation model	
	for(i=0; i<PLANES_NUMBER; i++) {
		//	Scheduling the first batch of departures
		sprintf(id, "%s-%d", cnames[LP], i);
		//	Sending the departure events
		TS_Schedule(clock+WAIT_T+i, (void *)create_submsg ("DEP", cnames[LP], id), sizeof(SubMsgHeader) );
	}
	
	//	 Main simulation loop, receives messages and calls the handler associated with them
	while (!end_reached) {

		// 	Looking for a new incoming message
		size = TS_Receive(&from, &Ts, (void *)msg, 1024);

		if(size > 0) {	
			// 	A message has been received, update the current simulated time
			//		and call the appropriate message handler
			submsg = (SubMsgHeader *)msg;

			//	Handlers
			if (strcmp (submsg->event_type, "ARR")==0)				// arrival event
				arrival_event_handler (from, submsg->from, submsg->id);
			
			if (strcmp (submsg->event_type, "LAN")==0)				// landing event
				landed_event_handler (submsg->from, submsg->id);
			
			if (strcmp (submsg->event_type, "DEP")==0){				// departure event
				total_flies++;
				departure_event_handler (submsg->id);
			}
		}
		else {	
			//	No more messages
			if (clock == end_clock) {	// checking if the simulation has finished

				printf ("[%s]: -- ending condition satisfied, clock=%.2f\n", cnames[LP], clock);
				printf ("Total flights departed from %s: %d\n", cnames[LP], total_flies);
    				end_reached = 1;

				//	The simulation is ended, we have to finalize the synchronization mechanism
				TS_Finalize();
			}
			else
				//	Still not reached the end_clock
				//		we have to advance to the next timestep
				clock = TS_TimeAdvance();
		}
	}
	
	return 0;
}
/*---------------------------------------------------------------------------*/
