/*	##############################################################################################
	Advanced RTI System, ARTÌS			http://pads.cs.unibo.it

	Description:	
		Airports example. 3 LP that model 3 interconnected airports. 
		In this model the synchronization mechanism is: Time Warp

	Output:
		the mode output can be found in the <LP_ID>.out file
		the middleware output is put on screen

	Changelog:
		13/10/09
			A few more comments

	Bugs:
		-	The current Time Warp implementation is quite unstable,
			some future debugging is required

	Authors:
		First version by Michele Bracuto <bracuto@cs.unibo.it>
		Modified by Gabriele D'Angelo <gda@cs.unibo.it>

	############################################################################################## */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <unistd.h>

//	ARTÌS - Time Warp header file
#include <tw.h>			

//	Model-related constants
#define FLIGHT_T		4	// time required to flight from an airport to another
#define LOAD_T			2	// time required to load/unload passengers
#define LAND_T			1	// landing time, required to free the runway
#define WAIT_T			1	// time a plane has to wait before re-requesting landing
#define PLANES_NUMBER		2	// planes initially in the hangar
//	Simulator-related constants
#define	END_SIMULATION		1	// end of simulation
#define	GVT_THRESHOLD		500	// global-virtual-time threshold (in timesteps)


/* ------------------SIMULATOR STATE VARIABLES-----------------*/

//	State of the simulator
typedef struct {
	int	runway_isfree;
	int	on_the_ground;
	int	in_the_air;
	int	total_flies;
	off_t 	offset;
	double	clock;
	double	SEED;
} _state_;

//	State pointers
//		model current state
_state_		_Pinfo_;
_state_		*Pinfo = &(_Pinfo_);
//		model state termination pointer
_state_		_PinfoTerm_;
_state_		*PinfoTerm = &(_PinfoTerm_);
//		termination pointer
int _Pend_reached_;
int *Pend_reached = &(_Pend_reached_);


/*------------------GLOBAL STATE VARIABLES------------------------*/

//	Time management variables
short	end_reached	= 0;		
double	end_clock	= 2000.0;			//	End of the simulated (simulated time)


/* ------------------SIMULATOR STATE VARIABLES-----------------*/

int	NEIGHBOUR1, NEIGHBOUR2, LP;			// identifiers
char	*cnames[3]={"MILAN", "BOLOGNA", "ROME"};	// aiport names


/*-----------MESSAGE STRUCTURE AND CONSTRUCTOR-----------*/

//	Structure of messages
typedef struct _sub_msg_header {
	char   event_type[4];	// type of message (model related)
	char   from[10];	// source of the message (airport name)
	char   id[12];		// flight identifier
} SubMsgHeader;

/*-------------------MESSAGE CREATION-----------------------------*/
char 	*create_submsg (char type[], char from[], char *id) {
	SubMsgHeader  * tmsg;

	tmsg = (SubMsgHeader *)malloc (sizeof(SubMsgHeader));
	memset (tmsg, 0, sizeof(SubMsgHeader));
	strcpy (tmsg->event_type, type);
	strcpy (tmsg->from, from);
	strcpy (tmsg->id, id);

	return (char *)tmsg;
}




/*---------------------------RANDOM GENERATOR-------------------*/

//	Seeds used by the generator
double SEEDS[20] = {
        606112567, 2005261598, 975679233, 381883932, 1949009736,
        457621696, 31040274, 2084138466, 63789657, 521636194,
        670018843, 1263912019, 2072256297, 1025373591, 1693757298,
        1115383532, 44685244, 1568381214, 2014569565, 1052870108
	};

//	Random numbers generator
static double my_rand ()
{
	static long m = 2147483647;
	static long a = 48271;
	static long q = 44488;
	static long r = 3399;
	long t, lo, hi;

	hi 	= Pinfo->SEED / q;
	lo 	= Pinfo->SEED - q * hi;
	t 	= a * lo - r * hi;

	if (t > 0)
		Pinfo->SEED = t;
	else
		Pinfo->SEED = t + m;


	return ( (double) Pinfo->SEED / (double) m );
}

//	Returns 0 or 1
static int	rnd_bool (void)
{
	return ( my_rand() <= 0.5 ? 0 : 1 );
}


/*--------------------------------- UTILS -----------------------------------*/

// TODO:
void	_printf(const char *fmt, ...)
{
	int	size;
	va_list	li;
	char	buf[1024];
		
	
	va_start(li, fmt);
	vsprintf(buf, fmt, li);
	va_end(li);

	size = strlen(buf);
	fseek(stdout, Pinfo->offset, SEEK_SET);

	fputs(buf, stdout);
	Pinfo->offset += size;
}
//
#define	printf	_printf




/*----------------------------EVENT HANDLERS------------------------*/

//	Arrival (to the local LP) of an airplane
void	arrival_event_handler (int src, char *from, char *id)
{

	//	Check the runway
	if (Pinfo->runway_isfree){
		//	It is free
		Pinfo->in_the_air++;		// one more airplane is waiting for landing
		Pinfo->runway_isfree = 0;	// lock the runway

		TW_Schedule(Pinfo->clock+LAND_T, (void *)create_submsg ("LAN", from, id), sizeof(SubMsgHeader));
		printf ("[%s]:   %.1f Waiting for landing, flight %s from %s\n", cnames[LP], Pinfo->clock, id, from);
		fflush(stdout);
	}
	else {
		//	If the source of the plane is not the local airport, then 
		//	one more airplane is waiting for landing	
		if(src != LP)	Pinfo->in_the_air++;
		
		//	Schedule a new arrival event in future
		TW_Schedule(Pinfo->clock+WAIT_T, (void *)create_submsg ("ARR", from, id), sizeof(SubMsgHeader));
		printf ("[%s]:   Runway busy, the flight %s from %s is waiting\n", cnames[LP], id, from);
		fflush(stdout);
	}	
}

//	Landing (to the local LP) of an airplane
void	landed_event_handler (char *da, char *id)
{
	//	One less on air, one more on the ground
	Pinfo->in_the_air--;
	Pinfo->on_the_ground++;

	//	Schedule the next departure (of the plane)
	TW_Schedule (Pinfo->clock + LOAD_T, (void *)create_submsg ("DEP", "", id), sizeof(SubMsgHeader));
	
	
	printf ("[%s]:     %.1f  Landed flight %s from %s\n", cnames[LP], Pinfo->clock, id, da);
	fflush(stdout);

	//	Now the runway is free
	Pinfo->runway_isfree = 1;
}

//	Departure (from the local LP) of an airplane
void	departure_event_handler (char *id)
{
	int dest;	// destination airport (LP)
	int ret;	// return value
	
	//	One less on the ground
	Pinfo->on_the_ground--;
	
	//	Choose a random destination from the 2 available
	dest = rnd_bool() == 0 ? NEIGHBOUR1 : NEIGHBOUR2;	
	
	//	Send to the destination airport (LP) the arrival event
	ret = TW_Send(dest, Pinfo->clock + FLIGHT_T, (void *)create_submsg ("ARR", cnames[LP], id), sizeof(SubMsgHeader) );
	
	//	Check the sending operation
	if(ret >= 0) {
		//	If OK then trace out the departure
		printf ("[%s]: %.1f Departed flight %s with destination %s\n",  cnames[LP], Pinfo->clock, /*clock + FLIGHT_T,*/ id, cnames[dest]);
		fflush(stdout);
	}
	else {
		//	Else print an error message
		//	FIXME
	}
}


/*----------------------------MAIN FUNCTION----------------------------*/

int main(int argc, char* argv[])
{
	SubMsgHeader  	*submsg;	// event message

	char	msg[1024];		// message buffer
	char 	id[20];			// airplane identification string
	int	from;			// source LP (of a message)
	int 	i;			// temporary variable
	long	ret;			// return value
	double	Ts;			// time variable
	char	out[100]; 		// Output buffer
	
	
	//	Setup the TIME WARP synchronization algorithm
	LP = TW_Init(cnames[atoi(argv[1])], argv[2], 5000, (void *)Pinfo, sizeof(*Pinfo), end_clock, GVT_THRESHOLD);

	// Redirection of standard output and error
	sprintf(out, "%d.out", LP);
	fclose(stdout); stdout = fopen(out, "w"); fflush(stdout);	
	
	//	Initialize the airports identifiers
	NEIGHBOUR1	= (LP + 1) % 3;
	NEIGHBOUR2	= (LP + 2) % 3;
	printf(" %s <- [%s] -> %s\n", cnames[NEIGHBOUR1], cnames[LP], cnames[NEIGHBOUR2] );

	//	Initialization of the current LP state		
	Pinfo->offset	= 0;
	Pinfo->SEED		= SEEDS[LP];		// random numbers generator, seed initialization
	Pinfo->clock		= 0.0;			// simulated time initialization

	//	Model variables initialization
	Pinfo->runway_isfree 	= 1;	// the runway is free
   	Pinfo->on_the_ground 	= 2;	// 2 planes (in each airport) are on the ground
   	Pinfo->in_the_air 	= 0;	// no airplanes are on the air
	Pinfo->total_flies	= 0;	// total number of departed flies (from this airport)
	
	//	Boostrapping the simulation model
	for(i=0; i<PLANES_NUMBER; i++) {
		//	Scheduling the first batch of departures
		sprintf(id, "%s-%d", cnames[LP], i);
		//	Sending the departure events
		TW_Schedule(Pinfo->clock+WAIT_T+i, (void *)create_submsg ("DEP", cnames[LP], id), sizeof(SubMsgHeader) );
	}
	
	//	 Main simulation loop, receives messages and calls the handler associated with them
	while (!end_reached) {

		// 	Looking for a new incoming message
		ret = TW_Receive( (void *)msg, &Ts, &from);

		if(ret != END_SIMULATION) {	
			
			// 	A message has been received, update the current simulated time
			//		and call the appropriate message handler
			submsg = (SubMsgHeader *)msg;

			//	Updating the simulated time of the current state
			Pinfo->clock = Ts;
			
			//	Handlers
			if (strcmp (submsg->event_type, "ARR")==0)				// arrival event
				arrival_event_handler (from, submsg->from, submsg->id);
			
			if (strcmp (submsg->event_type, "LAN")==0)				// landing event
				landed_event_handler (submsg->from, submsg->id);
			
			if (strcmp (submsg->event_type, "DEP")==0){				// departure event
				Pinfo->total_flies++;
				departure_event_handler (submsg->id);
			}
		}
		else {	
			//	No more messages
			printf ("[%s]: -- ending condition satisfied, clock=%.2f\n", cnames[LP], Pinfo->clock);
			printf ("Total flights departed from %s: %d\n", cnames[LP], Pinfo->total_flies);
			ftruncate(1, Pinfo->offset);

			//	The simulation if finished
			end_reached = 1;

			//	The simulation is ended, we have to finalize the synchronization mechanism
			TW_Finalize();
		}
	}
	
	return 0;
}
