/*	##############################################################################################
	Advanced RTI System, ARTÌS			http://pads.cs.unibo.it

	Description:
		HelloWorld example of using ARTÌS and GAIA with Java. 
		Many functionalities are missing, in this case the goals is to demonstrate that
		the GAIA APIs are usable with Java.

	Output:
		the output can be found in the following files:
		<LP_ID>.out	model output
		<LP_ID>.err	middleware output

	Changelog:
		24/09/10
			Added the GetStatistics() call
		13/10/09
			A little of commenting

	Authors:
		First version by Gabriele D'Angelo <gda@cs.unibo.it>

	############################################################################################## */

import java.math.BigDecimal;
import java.math.MathContext;

class HelloGaia {

	//	The simulation ends at this simulated time
	final static int	END_CLOCK = 20;

	//	Main procedure	
	public static void main(String args[]) {

		MathContext	mm = new MathContext(5);
		BigDecimal	la = new BigDecimal(1, mm);
		BigDecimal	clock = new BigDecimal(0, mm);
		//	LP identifier		
		int LP_id;
		//	Number of LPs	
		int nLP=0;
		//	Number of Simulated Entities (SEs) to simulate
		int nSim=0;

		// 	Create class instance
		ARTISjava javaOBJ=new ARTISjava();

  		//	Get as a parameter the number of Simulated Entities (SE) to simulate in each LP
		nSim	= Integer.parseInt(args[0]);
  		//	Get as a parameter the number of LPs in the simulation
		nLP	= Integer.parseInt(args[1]);
		//	Initialize the synchronization mechanism and get the local LP identifier
		LP_id	= javaOBJ.GAIA_Initialize(nSim*nLP, nLP, "Rand.seed", "", "localhost", 5000);
		
		//	Set the first identifier of the SEs to be simulated in the local LP
		//		remember that each SE has to be tagged with a unique identifier!
		javaOBJ.GAIA_SetFstID (nSim*LP_id);

		//	Turn ON the Migration functionalities of GAIA+
		javaOBJ.GAIA_SetMigration (ARTISjava.MIGR_ON);

		//	Tuning the migration factor, a float value is required by GAIA			
		javaOBJ.GAIA_SetMF ((float)1.1);
	
		// 	Turn ON the Load Balancing in GAIA+ (Default OFF)
		//		note:	ARTÌS Limited Edition does NOT support Load Balancing
//		javaOBJ.GAIA_SetLoadBalancing (ARTISjava.LOAD_ON);

		//	Register to GAIA all the [virtually] created SEs
		for(int i=0; i<nSim; i++) {

			javaOBJ.GAIA_Register(javaOBJ.MIGRABLE);
		}
	
		//	Main simulation cycle
		while (clock.doubleValue() < END_CLOCK) {

			//	Receive a new message (if available)	
			String data = new String(javaOBJ.GAIA_Receive(100));

			//	Checking the message type is possible to call the right handler	
			switch (javaOBJ.type) {
				
				//	Registration of new Simulated Entities (SEs) that are manager by other LPs
				case ARTISjava.REGISTER: 
					System.out.println("["+clock+"] --- REG.  ID " + javaOBJ.from + "  ON " + javaOBJ.to);
					break;

				// 	The GAIA framework informs the local LP the the migration of a 
				//	locally managed SE has to be done.
				//	The execution of the migration has not to be completed immediately
				//	but just before the end of the current timestep
				case ARTISjava.NOTIF_MIGR:
					System.out.println("["+clock+"] --- NOT.  ID " + javaOBJ.from + "  LP " + javaOBJ.to);
					break;

				//	A migration has been executed in the simulation.
				//	The local LP is not directly involved in the migration
				case ARTISjava.NOTIF_MIGR_EXT:
					System.out.println("["+clock+"] --- EXT.  ID " + javaOBJ.from + "  LP " + javaOBJ.to);
					break;

				//	The local LP is the recipient of a migration.
				//	The migration is implemented as a message with type "EXEC_MIGR" and
				//	containing the state of the migrating SE
				case ARTISjava.EXEC_MIGR:
					System.out.println("["+clock+"] --- MIGR.  ID " + javaOBJ.from + "  LP " + javaOBJ.to);
					break;
				
				//	End Of Step (EOS)
				case ARTISjava.EOS:
					System.out.println("["+clock+"] --- EOS\n");
					//	Send a broadcast message to all LPs in the simulation
					javaOBJ.GAIA_Broadcast(LP_id, clock.add(la).doubleValue(), "PING", 4);
					//	Getting some statistics from GAIA
					//	NOTE:	given that in this example there are only broadcasts
					//		it will be printed only zeroes
					javaOBJ.GAIA_GetStatistics();
					System.out.println("["+clock+"] --- Statistics, local: " + javaOBJ.local + " remote: " + javaOBJ.remote + " migrations: " + javaOBJ.migrations);
					//	Proceed with the next timestep
					clock = new BigDecimal(javaOBJ.GAIA_TimeAdvance(), mm);
					break;

				//	Is is a MODEL EVENT, that is not related to GAIA	
				case ARTISjava.UNSET:
					System.out.println("["+clock+"] --- From " + javaOBJ.from + " " + data);
					break;
				
				default:
					// Nothing
			}
		}

		//	The simulation is finished	
		System.out.println("["+clock+"] --- LP "+LP_id+" Finished!");

		//	Finalize the synchronization mechanism
		javaOBJ.GAIA_Finalize();
	}
}
