/*	##############################################################################################
	Advanced RTI System, ARTÌS			http://pads.cs.unibo.it

	Description:
		-	In this file you find all the user event handlers to be used to implement a 
			discrete event simulation. Only the modelling part is to be inserted in this 
			file, other tasks such as GAIA-related data structure management are 
			implemented in other parts of the code.
		-	Some "support" functions are also present.

	Authors:
		First version by Gabriele D'Angelo <gda@cs.unibo.it>

	############################################################################################### */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>
#include <ini.h>
#include <ts.h>
#include <rnd.h>
#include <gaia.h>
#include <rnd.h>
#include "utils.h"
#include "msg_definition.h"
#include "user_event_handlers.h"


/* ************************************************************************ */
/* 			E X T E R N A L     V A R I A B L E S 	            */
/* ************************************************************************ */

extern hash_t		hash_table, *table;		/* Global hash table of simulated entities */
extern hash_t		sim_table, *stable;		/* Hash table of locally simulated entities */
extern double		simclock;			/* Time management, simulated time */
extern TSeed		Seed, *S;			/* Seed used for the random generator */
extern int		LPID;				/* Identification number of the local Logical Process */
extern int		NSIMULATE;	 		// Number of Interacting Agents (Simulated Entities) per LP
extern int		NLP; 				// Number of Logical Processes
extern int		local_pid;			/* Process identifier */


/* ************************************************************************ */
/* 		 L O C A L	V A R I A B L E S			    */
/* ************************************************************************ */

// Total number of sent and received pings in this LP
unsigned long	lp_total_sent_pings 	= 0;
unsigned long	lp_total_received_pings = 0;


/* ************************************************************************ */
/* 		 S U P P O R T     F U N C T I O N S			    */
/* ************************************************************************ */

/* ***************************** D E B U G **********************************/

/*
	Prints out the whole content of a glib hashtable data structure
*/
void	UNUSED hash_table_print (GHashTable* ht) {

	// Iterator to scan the whole state hashtable of entities
	GHashTableIter		iter;
	gpointer		key, value;


	g_hash_table_iter_init ( &iter, ht );

	while ( g_hash_table_iter_next ( &iter, &key, &value ) ) {

		fprintf(stdout, "DEBUG: %d:%d\n", *(unsigned int *) key, *(unsigned int *) value);
		fflush(stdout);
	}	
}


/*
	Returns a random key from an hash table
*/
gpointer UNUSED hash_table_random_key (GHashTable* ht) {

	// Iterator to scan the (whole) state hashtable of entities
	GHashTableIter		iter;
	gpointer		key, value;
	//
	guint			size;
	unsigned int		position;


	size = g_hash_table_size( ht );
	position = RND_Integer( S, (double) 1, (double) size );
	
	g_hash_table_iter_init ( &iter, ht );

	while (position ) {

		g_hash_table_iter_next (&iter, &key, &value);
		position--;
	}	

	return(key);	
}


/*
	Statistics: exports the total number of SENT ping messages
*/
unsigned long	get_total_sent_pings () {

	return(lp_total_sent_pings);
}


/*
	Statistics: exports the total number of RECEIVED ping messages
*/
unsigned long	get_total_received_pings () {

	return(lp_total_received_pings);
}


/* *********** E N T I T Y    S T A T E    M A N A G E M E N T **************/

/*
	Adds a new entry in the hash table that implements the SE's local state
	Note: it is used both from the register and the migration handlers
*/
int	add_entity_state_entry (unsigned int key, value_element *val, hash_node_t *node) {

	struct	state_element	*state_e;


	// First of all, it is necessary to check if the used key is already in the hash table
	if ( g_hash_table_lookup ( node->data->state, &key ) != NULL)	return (-1);

	// The number of state records is limited by the MAX_MIGRATION_DYNAMIC_RECORDS constant,
	//	that is the max number of records that can be inserted in a migration message
	if ( g_hash_table_size( node->data->state ) > MAX_MIGRATION_DYNAMIC_RECORDS ) {

		// No more entries can be added, the resulting state would be impossible to migrate
		fprintf(stdout, "%12.2f node: FATAL ERROR, [%5d] impossible to add new elements to the state hash table of this node, see constant MAX_MIGRATION_DYNAMIC_RECORDS in file: sim-parameters.h\n", simclock, node->data->key);
		fflush(stdout);
		exit(-1);
	}

	// Dynamic allocation of memory and initialization of values
	// Note: this memory will be automatically freed in case of SE migration
	state_e = g_malloc( sizeof( struct state_element ) );

	if ( state_e ) {

		// Updating the state
		state_e->key		= key;
		state_e->elements	= *val;

		g_hash_table_insert (node->data->state, &(state_e->key), &(state_e->elements));

		#ifdef DEBUG
		fprintf(stdout, "%12.2f node: [%5d] local state key: %d, value: %d, local hash_size: %d\n", simclock, node->data->key, state_e->key, state_e->elements.value, g_hash_table_size (node->data->state));
		fflush(stdout);
		#endif

		return (1);
	}
	else {
		// Unable to allocate memory for state elements
		fprintf(stdout, "%12.2f node: FATAL ERROR, [%5d], memory allocation, impossible to add new elements to the state hash table of this node\n", simclock, node->data->key);
		fflush(stdout);
		exit(-1);
	}
}

/*
	Deletes an entry in the hash table that implements the SE's local state
	Note: the freeing of the associated memory is automatic
*/
int	delete_entity_state_entry (unsigned int key, hash_node_t *node) {

	if (g_hash_table_remove (node->data->state, &key) == TRUE)	return (0);
	else								return (-1);
}

/*
	Modifies the value of an entry in the SE's local state
*/
int	modify_entity_state_entry (unsigned int key, unsigned int new_value, hash_node_t *node) {

	unsigned int	*value;


	value = g_hash_table_lookup ( node->data->state, &key );

	if (value) {
		*(value) = new_value;
		return (0);
	} else	
		return (-1);
}


/* ********************** I N T E R A C T I O N S ***************************/

/*
  	Ping another SE, creating and sending a 'P' type message
	Usually called by user_generate_interactions_handler ()
 */
void	execute_ping (double ts, hash_node_t *src, hash_node_t *dest) 
{
	PingMsg		msg;
	unsigned int	message_size;
	int		tmp;


	// Defining the message type
	msg.ping_static.type = 'P';

	// The number of records in the ping message is chosen at random
	msg.ping_static.dyn_records = RND_Interval( S, 0, (double) MAX_PING_DYNAMIC_RECORDS );

	// A FAKE initialization for the records in the ping message
	for ( tmp = 0; tmp < msg.ping_static.dyn_records; tmp++ ) {

		msg.ping_dynamic.records[tmp].key	= tmp;
		msg.ping_dynamic.records[tmp].value	= tmp;
	}

	// To reduce the network overhead, only the used part of the message is really sent
	message_size = sizeof ( struct _ping_static_part ) + ( msg.ping_static.dyn_records * sizeof ( struct _ping_record ) );

	// Dynamic boudaries check
	if ( msg.ping_static.dyn_records > MAX_PING_DYNAMIC_RECORDS ) {

		fprintf(stdout, "%12.2f FATAL ERROR, the number of entries in the dynamic part of the ping message is bigger than MAX_PING_DYNAMIC_RECORDS!\n", simclock);
		fflush(stdout);
		exit(-1);
	}

	// Buffer check
	if ( message_size > BUFFER_SIZE ) {

		fprintf(stdout, "%12.2f FATAL ERROR, the outgoing BUFFER_SIZE is not sufficient!\n", simclock);
		fflush(stdout);
		exit(-1);
	}

	// Real send
	GAIA_Send (src->data->key, dest->data->key, ts, (void *)&msg, message_size);

	// Statistics
	lp_total_sent_pings++;
}

/* ************************************************************************ */
/* 		U S E R   E V E N T   H A N D L E R S			    */
/* ************************************************************************ */

/****************************************************************************
	PING: Upon arrival of an echo request do "nothing" or print some log
*/
void	user_ping_event_handler (hash_node_t *node, int from, Msg *msg) {

	// Statistics	
	lp_total_received_pings++;

	#ifdef DEBUG
	int	tmp;

	fprintf(stdout, "%12.2f node: [%5d] received a ping from agent [%5d], dyn_records: %d\n", simclock, node->data->key, from, msg->ping.ping_static.dyn_records);

	for ( tmp = 0; tmp <  msg->ping.ping_static.dyn_records; tmp++ ) {
			
		fprintf(stdout, "%12.2f - record %d, key %d, value %d\n", simclock, tmp, msg->ping.ping_dynamic.records[tmp].key, msg->ping.ping_dynamic.records[tmp].value);
	}
	#endif
}


/*
 	REGISTER: a new SE (in this LP) has been created, now it is possibile to
	initialize its data structures (es. local state)
*/
void	user_register_event_handler (hash_node_t *node) {

	value_element		val;
	int			tmp;
	

	// Initializing the local data structures of the node		
	node->data->state = g_hash_table_new_full ( g_int_hash, g_int_equal, g_free, NULL );

	// Generating (at random) some state values for the node
	for ( tmp = 0; tmp < RND_Interval( S, 1, (double) 5 ); tmp++ ) {

		// FAKE initialization
		val.value = tmp;

		// Adding a new entry in the local state of the registering node
		//	first entry	= key
		//	second entry	= value
		//	note: no duplicates are allowed
		if ( add_entity_state_entry( tmp, &val, node ) == -1 ) {

			// Insertion aborted, the key is already in the hash table
			fprintf(stdout, "%12.2f node: FATAL ERROR, [%5d] key %d (value %d) is a duplicate and can not be inserted in the hash table of local state\n", simclock, node->data->key, tmp, tmp);
			fflush(stdout);
			exit(-1);
		}
	}
}


/*
	NOTIFY MIGRATION: a local SE will be migrated in another LP.
	This notification is reported to the user level but usually nothing has to be done.
*/
void	user_notify_migration_event_handler () {

	// Nothing to do
}


/*
	NOTIFY EXTERNAL MIGRATION: SEs that are allocated in other LPs are going to be migrated, this LP is notified of this
	update but the user level usually does not care of it
*/
void	user_notify_ext_migration_event_handler () {

	// Nothing to do	
}


/*
	MIGRATION: migration-event manager (the real migration handler)

	A new migration message for this LP has been received, the trasported SE has been created and inserted in
	the data structures. Now it is necessary to perform some user level tasks such as taking care of de-serializing 
	the SE's local state
*/
void	user_migration_event_handler (hash_node_t *node, Msg *msg) {

	int			tmp;


	// Initializing the local data structures of the node		
	node->data->state = g_hash_table_new_full(g_int_hash, g_int_equal, g_free, NULL);

	// The migration message contains the state of the migrating SE,
	//	after allocating space to locally manage the node, I've
	//	now to update the state of the SE using the state 
	//	information contained in the migration message
        //
        // Static part
        node->data->s_state = msg->migr.migration_static.s_state;
        // Dynamic part
	for ( tmp = 0; tmp < msg->migr.migration_static.dyn_records; tmp++ ) {

		// Adding a new entry in the local state of the migrated node (rebuilds its local state)
		//	first entry	= key
		//	second entry	= value
		//	note: no duplicates are allowed
		if ( add_entity_state_entry(msg->migr.migration_dynamic.records[tmp].key, 
			&(msg->migr.migration_dynamic.records[tmp].elements), node) == -1 ) {

			// Insertion aborted, the key is already in the hash table
			fprintf(stdout, "%12.2f node: FATAL ERROR, [%5d] key %d (value %d) is a duplicate and can not be inserted in the hash table of local state\n", simclock, node->data->key, msg->migr.migration_dynamic.records[tmp].key, msg->migr.migration_dynamic.records[tmp].elements.value);
			fflush(stdout);
			exit(-1);
		}
	}
}


/*
	CONTROL: at each timestep, the LP calls this handler to permit the execution of model level interactions,
	for performance reasons the handler is called once for all the SE that allocated in the LP	
*/
void	user_control_handler () {

	int		h, destination;
	hash_node_t 	*node, *dest;


	// For each local SE
   	for ( h = 0; h<stable->size; h++ ) {
		for ( node = stable->bucket[h]; node; node = node->next) {	

			// The destination SE is chosen at random
			destination = RND_Interval( S, 0, (double) NSIMULATE * NLP );

			// The chosen destination is verified against the global hash table
			if ( ( dest = hash_lookup ( table, destination ) ) )  {
		
				// The ping message will arrive to the received in FLIGHT_TIME timesteps
				execute_ping (simclock + FLIGHT_TIME, node, dest);
			}
			else {
				// I'm trying to send a ping to a node that does not exist in the global hash table
				fprintf(stdout, "%12.2f node: FATAL ERROR, [%5d] is NOT a valid destination!\n", simclock, destination);
				fflush(stdout);
				exit(-1);
			}
		}	
	}
}

/*
	USER MODEL: when it is received a model level interaction, after some validation this generic handler is called
	Now, the specific user level handler will complete its processing
*/
void	user_model_events_handler (int to, int from, Msg *msg, hash_node_t *node) {

	// A model event has been received, now calling appropriate user level handler
	switch (msg->type) {

		case 'P':	// Ping message
			user_ping_event_handler( node, from, msg );
		break;

		default:
			fprintf(stdout, "FATAL ERROR, received an unknown user model event type: %d\n", msg->type);
			fflush(stdout);
			exit(-1);
	}
}

/*
        BOOTSTRAP: before starting the real simulation tasks, the model level 
        can initialize some data structures and set parameters
*/
void    user_bootstrap_handler (unsigned long total_number_se, unsigned long number_lp, unsigned long local_number_se) {

	// Nothing to do
}


/*
	SHUTDOWN: Before shutting down, the model layer is able to 
	deallocate some data structures
*/
void	user_shutdown_handler () {

	// Nothing to do
}

