/*	##############################################################################################
	Advanced RTI System, ARTÌS			http://pads.cs.unibo.it

	Description:	
		GAIA+ enabled wireless example. It simulates a set of  Simulated Mobile Hosts 
		(SMHs) moving using the Random Way Point (RWP) mobility model in a toroidal 
		2D space. Some of the hosts (nodes) are equipped with a wireless device 
		(default=100%). The interactions among the (wireless) devices are based on 
		proximity.

	Notes:
		RWP:	min speed = max speed
		RWP:	other parameters are in wireless.ini
		
	Output:
		the mode output can be found in the <LP_ID>.out file
		the middleware output is put on screen

	Changelog:
		28/03/13
			Fixed a bug in the implementation of the Random Waypoint Mobility, the PAUSE time was not 	
			managed correctly. The model behavior was affected by the state of some uninitialized 
			variables.
		24/01/13
			Better definition of the migration ratio
		23/01/13
			Polishing, better comments and stats
		18/01/13
			Random Way Point mobility model polishing and implemented the sleep time of
			nodes (at bootstrap and after reaching a target destination).
		10/01/13
			Each simulated entity now has a local seed for random number generation,
				this seed is migrated among with the entity state and therefore
				the adaptive simulation (GAIA ON) generates very the same trace 
				than the non adaptive one (GAIA OFF).
			Fixed the mobility model implementation: in some cases the initial state 
				of a few simulated entities was not correctly broadcasted. 
				It was a minor issue given that in a few steps the movement 
				updates were able to obtain exactly the same data dissemination.
			Some code polishing.
		16/09/11
			Random way point bugfix: thanks to Serena Pirazzini
		31/08/10
			Some cosmetic changes
			Fixed some errors in the defines
			Added more detailed output and some runtime controls
		16/11/09
			Fixed some errors in comments
			Dynamic allocation of the receiving buffer
		10/11/09
			Added support for ping and migration extra payload
		26/10/09
			Cosmetic changes	

	Authors:
		First version by Michele Bracuto <bracuto@cs.unibo.it>
		Modified by Gabriele D'Angelo <g.dangelo@unibo.it>

	############################################################################################## */

/*	TODO
	
	1) nothing
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>
#include <ini.h>
#include <ts.h>
#include <rnd.h>
#include <gaia.h>

/*-------------------------- D E B U G --------------------------------------*/
// Uncomment to activate debug
//#define DEBUG

// Uncomment to activate the output trace used by the graphic viewer
//#define VISUALIZER

/*---------------------------------------------------------------------------*/

#define	UNUSED			__attribute__ ((__unused__))
#define ASSERT( _cond, _act )   { if( !(_cond) ) { printf _act; printf("\n"); assert( _cond ); } }

#define TIMER_NOW(_t) 		gettimeofday(&_t,NULL)
#define TIMER_SECONDS(_t) 	((double)(_t).tv_sec + (_t).tv_usec*1e-6)
#define TIMER_DIFF(_t1, _t2) 	(TIMER_SECONDS(_t1)-TIMER_SECONDS(_t2))

#define PI			3.1415926535897932384626433832795

#define MIN(X, Y)  		((X) < (Y) ? (X) : (Y))
#define MAX(X, Y)  		((X) > (Y) ? (X) : (Y))
#define	YES			'1'
#define	NO			'0'

//	The Simulated Mobile Host can be migrated or not?
#define	MIGRABLE		'1'
#define	NOT_MIGRABLE		'0'

/*---------------------------------------------------------------------------*/

// Proximity detection mechanism (all pairs algorithm)
//	to use squares optimization uncomment the define
//	for more details: http://doi.acm.org/10.1145/1164717.1164727
#define	USE_SQUARE

// Simulation length (final clock value)
#define END_CLOCK			1000.0

// Number of timestep required by ping messages to receive the destination node
//	WARNING: due to synchronization constraints The FLIGHT_TIME has to be bigger 
//		than the timestep size
#define FLIGHT_TIME			1.0

// Extra payload to be inserted in PING packets used as interactions among
// 	simulated entities. It is very useful to test the performance of
//	the simulation and migration mechanism in different conditions
#define	PING_PAYLOAD_EXTRA_SIZE		0

// Extra payload to be inserted in the MIGRATION of simulated entities,
// 	it is very useful to test the performance of 
//	the migration mechanism in different conditions
#define	MIGRATION_PAYLOAD_EXTRA_SIZE	0

// Buffer size for incoming messages
//	obviously the buffer needs to be so large to contain all kind of
//	messages (e.g. see the PING_PAYLOAD_EXTRA_SIZE and
//	MIGRATION_PAYLOAD_EXTRA_SIZE values)
#define BUFFER_SIZE			65535

// Mobility and wireless model parameters
//
//	probability (in this timestep) to send an interaction to a wireless host that is within 
//		the communication range of the simulated host
#define	PERC_PING			0.2
//	only a part of the simulated hosts is equipped with a wireless device
//		probability that the simulated hosts is wireless-enabled
//		(default 1.0 = 100% wireless enabled)
#define	PERC_COMM			1.0

/*---------------------------------------------------------------------------*/

static int      NSIMULATE, 		// Number of Simulated Mobile Hosts (Simulated Entities) per LP
		NLP, 			// Number of Logical Processes
		LPID;			// Identification number of the local Logical Process 

// Random-Waypoint mobility model, motion speed
static int	SPEED;

// Random-Waypoint mobility model, pause time
static int	PAUSE;

// Playground definition (X and Y size)
static float	MAX_X = 0.0, MAX_Y = 0.0;

// Communication range of wireless devices
static float	RADIUS;

// Proximity detection mechanism, squares optimization
// for more details: http://doi.acm.org/10.1145/1164717.1164727
#ifdef USE_SQUARE
static float	RADIUS_INT;
#endif

// SImulation MAnager information and localhost identifier
static char	LP_HOST[64];
static char	SIMA_HOST[64];
static int	SIMA_PORT;

// Time management variables
static double   step, 			// Timestep size
		simclock = 0.0;		// Simulated time
static int	end_reached = 0;	// Simulation exit condition

// A single LP is responsible to show the runtime statistics
//	by default the first started LP is responsible of this task
static int	LP_STAT = 0;
// In some cases (for performance reasons) a LP can exit from the simulation
//	before the predefined ending
static int	to_leave = 0;

// Total number of sent and received pings in this LP
static long	lp_total_sent_pings = 0;
static long	lp_total_received_pings = 0;

// File containing a set of seeds used by the random numbers generator
char		*rnd_file = "Rand.seed";

/*---------------------------------------------------------------------------*/

// Model messages definition
typedef struct _move_msg    	MoveMsg;
typedef struct _ping_msg    	PingMsg;
typedef struct _comm_msg    	CommMsg;
typedef struct _migr_msg    	MigrMsg;
typedef union   msg         	Msg;

// Movement message
struct _move_msg {
	char    type;					// message type
	float	x;					// X-coordinate
	float	y;					// Y-coordinate
};

// Ping message
struct _ping_msg {
	char   type;					// message type
	char   filler[PING_PAYLOAD_EXTRA_SIZE];		// filler to simulated extra payload
};

// State of the communication device
struct _comm_msg {
	char   type;					// message type
	char   comm_device;				// communication device ON/OFF
};

// Migration message
struct _migr_msg {
	char		type;					// message type
	unsigned char	rwp_state;				// random way point mobility state
	float		targX;					// X-coordinate
	float		targY;					// Y-coordinate
	TSeed		Seed;					// seed used for the random generator
	char		filler[MIGRATION_PAYLOAD_EXTRA_SIZE];	// filler to simulate extra payload	
};

// Union structure for messages
union msg {
	char	 type;
    	MoveMsg  move;
    	CommMsg  comm;
    	PingMsg  ping;
	MigrMsg  migr;
};

/*---------------------------------------------------------------------------*/

/* ************************************************************************ */
/* 			            Hash Tables		      	            */
/* ************************************************************************ */
enum HASH_TYPE {
	SMH,		/* Global hash table of simulated mobile hosts */
	SIM,		/* Hash table of locally simulated mobile hosts */
};

typedef struct hash_data_t {
	int		key;		// SMH identifier
	int 		lp;		// LP in which the SMH is run
	float		posX;		// current position X-axis
	float		posY;		// current position Y-axis
	float		targX;		// target position X-axis
	float		targY;		// target position Y-axis
	char		changed;	// used by graphic viewer
	char		mobile;		// static or dynamic simulated entity?
	char		comm_device;	// wireless enabled or not?
	TSeed		Seed;		// seed used for the random generator
	unsigned char	rwp_state;	// random way point mobility state
} hash_data_t;

typedef struct hash_node_t {
	struct hash_data_t *data;
   	struct hash_node_t *next;
} hash_node_t;

typedef struct hash_t {
   	struct hash_node_t **bucket;
	int count;
   	int size;
} hash_t;

hash_t		hash_table,  *table=&hash_table;	/* Global hash table of simulated mobile hosts */
hash_t		sim_table,   *stable=&sim_table;	/* Hash table of locally simulated mobile hosts */

/*---------------------------------------------------------------------------*/


/* ************************************************************************ */
/* 			   Migrating Objects List			    */
/* ************************************************************************ */
typedef struct smh_list_n {
	struct hash_node_t *node;
   	struct smh_list_n  *next;
} smh_list_n;


typedef struct smh_list {
   	struct smh_list_n *head;
	struct smh_list_n *tail;
	int size;
} smh_list;

// List of the objects (SE) that have to migrate at the end of the current timestep
static	smh_list	migr_list,	
	*mlist 	= 	&migr_list;

/*---------------------------------------------------------------------------*/


/* ************************************************************************* */
/* 	         D A T A    S T R U C T U R E S    M A N A G E M E N T       */
/* ************************************************************************* */

static int hash(hash_t *tptr, int x)
{
	return(x % tptr->size);
}
/*---------------------------------------------------------------------------*/


/*
	Hash table initialization
*/
static void hash_init(hash_t *tptr, int size)
{
   	tptr->size	= size;
	tptr->count	= 0;
   	tptr->bucket	= (hash_node_t **) calloc(tptr->size, sizeof(hash_node_t *));

   	return;
}


/*
	Lookup of a node (hash table)
*/
static hash_node_t *hash_lookup(hash_t *tptr, int key)
{
   	int 		h;
	hash_node_t 	*node;


   	h = hash(tptr, key);
   	for ( node=tptr->bucket[h]; node != NULL; node = node->next ) {
     		if ( node->data->key == key )
      			break;
   	}

   	return(node);
}


/*
	Insertion of a new node (hash table)
*/
static hash_node_t * hash_insert(enum HASH_TYPE type, hash_t *tptr, struct hash_data_t *data, int key, int lp)
{
   	hash_node_t 	*node, *tmp;
   	int 		h;

	
	if ( ( tmp = hash_lookup(tptr, key) ) )
		return(tmp);
	
	h = hash(tptr, key);

	node = (struct hash_node_t *) malloc(sizeof(hash_node_t));
	ASSERT(( node != NULL ), ("hash_insert: malloc error"));

	if(type == SMH) {
		node->data = (struct hash_data_t *) malloc(sizeof(hash_data_t));
		ASSERT(( node->data != NULL ), ("hash_insert: malloc error"));

		node->data->key	   	= key;
		node->data->lp	   	= lp;
		node->data->posX	= 0.0;
		node->data->posY	= 0.0;
	}
	else if(type == SIM) {
		node->data 		= data;
		node->data->lp	   	= lp;
	}

   	node->next	= tptr->bucket[h];
   	tptr->bucket[h]	= node;
        tptr->count	+= 1;
		
   	return node;
}


/*
	Deletion of a node (hash table)
*/
static int UNUSED hash_delete(enum HASH_TYPE type, hash_t *tptr, int key)
{
	hash_node_t	*node, *last;
   	int h;

	
   	h = hash(tptr, key);
   	for ( node = tptr->bucket[h]; node; node = node->next ) {
     		if ( node->data->key == key )
       			break;
   	}

   	if ( node ==NULL )
     		return -1;

   	if ( node == tptr->bucket[h] )
     		tptr->bucket[h] = node->next;
   	else {
     		for ( last= tptr->bucket[h]; last && last->next; last = last->next ) {
       			if ( last->next == node )
         			break;
     		}
     		last->next = node->next;
   	}

	tptr->count -= 1;
	
	if( type == SMH )
		free(node->data);

	free(node);

   	return(1);
}


/*
	List initialization
*/
static void	list_init (smh_list *list)
{
	list->head	= NULL;
	list->tail	= NULL;
	list->size	= 0;
}
/*---------------------------------------------------------------------------*/


/*
	Insertion of a new node (list)
*/
static void	list_add (smh_list  *list, hash_node_t *node)
{
	smh_list_n 	*list_n;

	
   	list_n		  = (struct smh_list_n *) malloc(sizeof(smh_list_n));
	ASSERT(( list_n != NULL ), ("list_add: malloc error"));

   	list_n->node	  = node;
   	list_n->next	  = NULL;

	if ( list->head == NULL ) {
		list->head = list_n;
	}

	
	if ( list->tail != NULL ) {
		list->tail->next = list_n;
	}

	list->tail 	 = list_n;
	list->size	+= 1;
}


/*
	Deletion of a node (list)
*/
struct hash_node_t *list_del (smh_list  *list) 
{
	struct smh_list_n	*head = NULL;
	struct hash_node_t	*node = NULL;

	
	if( list->size > 0 ) {

		list->size     -= 1;

		head 		= list->head;
		node 		= list->head->node;
		list->head	= list->head->next;
		
		if ( list->size <= 0 ) 
			list->tail = NULL;
			
		free(head);
	}
	

	return node;
}
/*---------------------------------------------------------------------------*/


/* ************************************************************************ */
/* 		      M O D E L    D E F I N I T I O N			    */
/* ************************************************************************ */

/*
	Random WayPoint (RWP) mobility model
*/
static void UNUSED	 rwp_move(hash_node_t *smh_node)
{
	float	tmpX, tmpY;
        double 	angle;

	// If the node is approaching the target destination
	if ( ( fabs( smh_node->data->posX - smh_node->data->targX ) < SPEED ) && ( fabs( smh_node->data->posY - smh_node->data->targY ) < SPEED ) ) {

		#ifdef DEBUG
		fprintf(stdout, "%12.2f node: [%5d] is approaching ([%5.2f][%5.2f]) its target destination [%5.2f][%5.2f]\n", simclock, smh_node->data->key, smh_node->data->posX, smh_node->data->posY, smh_node->data->targX, smh_node->data->targY);
		#endif
			
		// Now the node is exactly on the destination spot
		smh_node->data->posX = smh_node->data->targX;
		smh_node->data->posY = smh_node->data->targY;

		// A new target destination has to be chosen
		smh_node->data->targX	= (float)RND_Interval (&(smh_node->data->Seed), 0, MAX_X);
		smh_node->data->targY	= (float)RND_Interval (&(smh_node->data->Seed), 0, MAX_Y);
		
		#ifdef DEBUG
		fprintf(stdout, "%12.2f node: [%5d] has reached its target destination, new destination [%5.2f][%5.2f]\n", simclock, smh_node->data->key, smh_node->data->targX, smh_node->data->targY);
		#endif	
			
		// Has the node to be paused for some amount of time?
		if (PAUSE) {
			
			smh_node->data->rwp_state = (int)RND_Interval(&(smh_node->data->Seed), 0, PAUSE);
			#ifdef DEBUG
			fprintf(stdout, "%12.2f node: [%5d] will pause for [%3d] timesteps\n", simclock, smh_node->data->key, smh_node->data->rwp_state);
			#endif
		} else {
			smh_node->data->rwp_state = 0;
		}
	} else {
		// In the following we need to update the node position	
		
		tmpX = smh_node->data->targX;
		tmpY = smh_node->data->targY;

		// Shortest path determination
		if ( fabs( smh_node->data->targX - smh_node->data->posX ) > MAX_X/2 ) {
			if ( smh_node->data->targX > smh_node->data->posX )
				tmpX	-= MAX_X;
			else
				tmpX	+= MAX_X;
		}

		if ( fabs(smh_node->data->targY - smh_node->data->posY ) > MAX_X/2 ) {
			if ( smh_node->data->targY > smh_node->data->posY )
				tmpY	-= MAX_X;
			else
				tmpY	+= MAX_X;
		}

		angle	= atan( ( tmpY - smh_node->data->posY ) / ( tmpX - smh_node->data->posX ) );

		if ( ( tmpX - smh_node->data->posX < 0 ) ) {
			angle += PI;
		}

		tmpX	= smh_node->data->posX + (SPEED * cos(angle));
		tmpY	= smh_node->data->posY + (SPEED * sin(angle));

		if ( tmpX >= MAX_X)	{tmpX -= MAX_X;}
		if ( tmpY >= MAX_Y)	{tmpY -= MAX_Y;}
		if ( tmpX < 0 )		{tmpX += MAX_X;}
		if ( tmpY < 0 )		{tmpY += MAX_Y;}

		smh_node->data->posX	= tmpX;
		smh_node->data->posY	= tmpY;
		
		#ifdef DEBUG
		fprintf(stdout, "%12.2f node: [%5d] new position [%5.2f][%5.2f]\n", simclock, smh_node->data->key, smh_node->data->posX, smh_node->data->posY);
		#endif
	}
}

/*
	Random generation of initial position
*/
static void	rwp_initial_position (float *posX, float *posY, TSeed *Seed)
{
	float	tmpx, tmpy;

	
	tmpx  	= (float)RND_Interval (Seed, 0, MAX_X); 
	tmpy  	= (float)RND_Interval (Seed, 0, MAX_Y); 

	*posX = tmpx;
	*posY = tmpy;
}


/*
	SMH Movement
*/
static void	move_event (double ts, hash_node_t *node)
{
	MoveMsg		msg;
	int		paused = 0;

	
	// The node is paused by the mobility model (i.e. has just reached the target destination)
	if ( ( PAUSE ) && ( node->data->rwp_state > 0 ) ) {
		
		// reducing the pausing time
		(node->data->rwp_state)--;
		
		paused = 1;

		#ifdef DEBUG
		fprintf(stdout, "%12.2f node: [%5d] is paused -> [%3d] remaining timesteps\n", simclock, node->data->key, node->data->rwp_state);
		#endif
	}	

	// If the node is not paused the current position is updated
	if ( !paused ) {
		
		// Generating the new RWP position
		rwp_move(node);

		// A new position update message has to be broadcasted
		msg.type	= 'M';
		msg.x		= node->data->posX;
		msg.y		= node->data->posY;

		GAIA_Broadcast(node->data->key, ts, (void *)&msg, sizeof(MoveMsg));
	}
}


/*
	Movements generation
*/
static void	ScanActivity()
{
	int		h;
	hash_node_t 	*node;

	
   	for ( h = 0; h < stable->size; h++ ) {
		
		for ( node = stable->bucket[h]; node; node = node->next )
			// Subset of movable entities
			if ( node->data->mobile == YES ) 
				move_event(simclock + step, node);
	}
}


/*
	Euclidean Distance
*/
static float UNUSED distance(float diff_x, float diff_y)
{
	return( sqrtf( ( diff_x * diff_x ) + ( diff_y * diff_y ) ) );
}


/*
 * 	Ping another SMH, creating a 'P' type message and sending it to correct LP
 */
static void	ping(double ts, hash_node_t *src, hash_node_t *dest)
{
	PingMsg	msg;
	

	msg.type = 'P';

	GAIA_Send(src->data->key, dest->data->key, ts, (void *)&msg, sizeof(PingMsg));

	lp_total_sent_pings++;

}


/*
	Send a Ping to all SMHs (nodes) within the communication range
*/
static void UNUSED signal_to_range(hash_t *tptr, hash_node_t *smh_node)
{
	int	h;
	float 	dist;
	#ifdef USE_SQUARE
	float	distX, distY;
	#endif
	hash_node_t *peer;


	// It is necessary to scan the whole hashtable containing the simulated entities
   	for ( h = 0; h < tptr->size; h++ ) {
		
		for ( peer = tptr->bucket[h]; peer; peer = peer->next ) {
			
			// It is not a good idea to ping itself
			if ( peer->data->key == smh_node->data->key  ) 
				continue;

			// If the sender has no wireless devices then no ping can be sent
			if( peer->data->comm_device == NO  ) 
				continue;
			
			// Proximity detection mechanism, squares optimization
			// for more details: http://doi.acm.org/10.1145/1164717.1164727
			#ifdef USE_SQUARE
			distX = fabsf( peer->data->posX-smh_node->data->posX );
			distY = fabsf( peer->data->posY-smh_node->data->posY );

			// Checks the external square
			if ( ( distX <= RADIUS ) && ( distY <= RADIUS ) ) {
				// Checks the internal square
				if ( ( distX <= RADIUS_INT ) && ( distY <= RADIUS_INT ) ) {
					
					#ifdef DEBUG
					fprintf(stdout, "%12.2f node: [%5d] is pinging node [%5d]\n", simclock, smh_node->data->key, peer->data->key);
					#endif

					// The ping message will arrive to the received in FLIGHT_TIME timesteps
					ping(simclock + FLIGHT_TIME, smh_node, peer);
				}
				else {
			#endif
					// Points that are placed in the area between the internal and the external square or
					// all points if square optimization is not enabled
					
					// Euclidean distance
					dist = distance(peer->data->posX - smh_node->data->posX, peer->data->posY - smh_node->data->posY);

					// Within the transmission range or not?
					if ( ( dist <= RADIUS ) ) {

						#ifdef DEBUG
						fprintf(stdout, "%12.2f node: [%5d] is pinging node [%5d]\n", simclock, smh_node->data->key, peer->data->key);
						#endif

						// The node is within the range
						ping(simclock + FLIGHT_TIME, smh_node, peer);
					}
			#ifdef USE_SQUARE
				}
			}
			#endif
		}
		
	}
}


/*
 	Sends a Ping to all SMHs that are inside the action range
*/
static void UNUSED PingRange()
{
	int		h;
	hash_node_t 	*node;


   	for ( h = 0; h < stable->size; h++) {
		for ( node = stable->bucket[h]; node; node = node->next ) {
			// If the sender has no wireless devices then no ping is sent
			if ( node->data->comm_device == NO  ) 
				continue;
			
			// To ping or not to ping? The evaluation is probability based
			if( RND_Unif(&(node->data->Seed)) < PERC_PING )
				signal_to_range(table, node);
		}
	}
}


/*
	SMH generation, called once when global variables have been initialized.
*/
static void Generate(int count)
{
	int	i;


	// All the simulated entities in this LP can be migrated
	for ( i = 0; i < count; i++ ) {
		
		GAIA_Register(MIGRABLE);
	}
}


/*
	Performs the migration of the flagged SMHs
*/
static int UNUSED ScanMigrating()
{
	struct hash_node_t	*smh=NULL;
	MigrMsg     		m;
	int 			migrated_in_this_step = 0;

	
	// The SMHs to migrate have been already identified by GAIA
	//	and placed in the migration list (mlist) when the
	//	related NOTIF_MIGR was received
	while ( ( smh = list_del(mlist) ) ) {

		migrated_in_this_step++;

		// A new "M" (migration) type message is created
		m.type	= 'M';
		// The state of the SMH has to be inserted in the migration message 
		m.targX	= smh->data->targX;
		m.targY = smh->data->targY;
		m.Seed = smh->data->Seed;
		m.rwp_state = smh->data->rwp_state;

		// The real migration is executed
		GAIA_Migrate(smh->data->key, (void *)&m, sizeof(MigrMsg));

		// Removing the migrated SMH from the local list of migrating nodes
		hash_delete(SIM, stable, smh->data->key);
	}

	return(migrated_in_this_step);
}
/*---------------------------------------------------------------------------*/


/* ************************************************************************ */
/* 			 E V E N T   H A N D L E R S			    */
/* ************************************************************************ */

/*
	Upon arrival of an echo request do "nothing" or print some log
*/
static void	ping_event_handler(int id, int to, Msg *msg)
{
	hash_node_t	*node;	

	if ( ( node = hash_lookup(stable, to) ) )  {

		#ifdef DEBUG
		fprintf(stdout, "%12.2f node: [%5d] received a ping from node [%5d]\n", simclock, to, id);
		#endif

		lp_total_received_pings++;
	} else {
		// The receiver is not managed by this LP, it is really a fatal error
		fprintf(stdout, "%12.2f node: FATAL ERROR, [%5d] is NOT in this LP!\n", simclock, to);
		fflush(stdout);
		exit(-1);
	}
}


/*
 	Updates the SMH position
*/
static void	move_event_handler(int id, Msg *msg)
{
	hash_node_t *node;

	
	if ( ( node = hash_lookup(table, id) ) ) {
	
		node->data->changed	= YES;

		node->data->posX = msg->move.x;
		node->data->posY = msg->move.y;
	}

}


/*
	Turns ON/OFF the wireless device of a SMH
		ON = the SMH is wireless enabled
		OFF = the SMH is NOT wireless enable
*/
static void	comm_event_handler(int id, Msg *msg)
{
	hash_node_t *node;

	
	if ( ( node = hash_lookup( table, id) ) ) {

		node->data->changed	= YES;
		node->data->comm_device = msg->comm.comm_device;
	}

}


/*
 	A new SMH has been created, we have to insert the SMH into the global 
	and local hashtables in the position given by sender's ID
*/
static void	register_event_handler(int id, int lp)
{
 	hash_node_t 	*node;
	CommMsg		c_msg;
	MoveMsg		m_msg;
	
	
	node = hash_insert(SMH, table, NULL, id, lp);
	
	if ( node ) {
		
		node->data->changed	= YES;
		node->data->mobile	= YES;
		node->data->comm_device	= NO;
		
		// If the SMH is local then it has to be inserted also in the local
		//	hashtable and some extra management is required	
		if ( lp == LPID ) { 
			// Inserting it in the table of local SMHs
			hash_insert(SIM, stable, node->data, node->data->key, LPID);

			// Initialization of the random numbers generator
			RND_Init(&(node->data->Seed), rnd_file, LPID + id);			
			
			// Random Way Point
			//	initial position
			rwp_initial_position(&(node->data->posX), &(node->data->posY), &(node->data->Seed));
			#ifdef DEBUG
			fprintf(stdout, "%12.2f node: [%5d] has initial position [%5.2f][%5.2f]\n", simclock, node->data->key, node->data->posX, node->data->posY);
			#endif
			
			//	initial state
			if (PAUSE)	node->data->rwp_state	= (int)RND_Interval(&(node->data->Seed), 0, PAUSE);
			#ifdef DEBUG
			fprintf(stdout, "%12.2f node: [%5d] has initial pause time [%3d]\n", simclock, node->data->key, node->data->rwp_state);
			#endif
			
			// A new position update message has to be broadcasted
			m_msg.type	= 'M';
			m_msg.x		= node->data->posX;
			m_msg.y		= node->data->posY;

			GAIA_Broadcast(node->data->key, simclock+step, (void *)&m_msg, sizeof(MoveMsg));
			
			// Generation of a new target position
			node->data->targX	= (float)RND_Interval(&(node->data->Seed), 0, MAX_X);
			node->data->targY	= (float)RND_Interval(&(node->data->Seed), 0, MAX_Y);

			#ifdef DEBUG
			fprintf(stdout, "%12.2f node: [%5d] has target position [%5.2f][%5.2f]\n", simclock, node->data->key, node->data->targX, node->data->targY);
			#endif
					
			// Management of the wireless device in the SMH
			//	only some SMHs are equipped with wireless devices
			if ( RND_Unif(&(node->data->Seed)) < PERC_COMM ) {
				c_msg.type		= 'W';
				c_msg.comm_device	= YES;
			
				// The information about the SMH's wireless device state has to be
				//	propagated to the whole simulaton
				GAIA_Broadcast(id, simclock+step, (void *)&c_msg, sizeof(CommMsg));
			}
		}
	}
}


/*
	Manages the "migration notification" of local SMH
	(that is, allocated in the local LP)
*/
static void	notify_migration_event_handler(int id, int to)
{
	hash_node_t *node;


	#ifdef DEBUG
	fprintf(stdout, "%12.2f node: [%5d] is going to be migrated in LP [%5d]\n", simclock, id, to);
	#endif
	
	// The GAIA framework has decided that a local SMH has to be migrated,
	//	the migration can NOT be executed immediately because the SMH
	//	could be the destination of some "in flight" messages
	if ( ( node = hash_lookup(table, id) ) )  {
		/* Now it is updated the list of SMHs that are enabled to migrate (flagged) */
		list_add(mlist, node);

		node->data->lp		= to;
		node->data->changed	= YES;
	}

	// Just before the end of the current timestep, the migration list will be emptied
	//	and the pending migrations will be completed
}


/*
	Manages the "migration notification" of external SMHs
	(that is, NOT allocated in the local LP)
*/
static void	notify_ext_migration_event_handler(int id, int to)
{
	hash_node_t *node;

	
	// A migration that does not directly involve the local LP is going to happen in
	//	the simulation. In some special cases the local LP has to take care of 
	//	this information 
	if ( (node = hash_lookup(table, id) ) )  {
		node->data->lp		= to;
		node->data->changed	= YES;
	}
}


/*
	Migration-event manager (the real migration handler)

	This handler is execute when a migration message is received and
	therefore a new SMH has to be accomodated in the local LP
*/
static void	migration_event_handler(int id, Msg *msg)
{
	hash_node_t 	*node;

	
	#ifdef DEBUG
	fprintf(stdout, "%12.2f node: [%5d] has been migrated in this LP\n", simclock, id);
	#endif

	if ( ( node = hash_lookup(table, id) ) ) {
		// Inserting the new SMH in the local table
		hash_insert(SIM, stable, node->data, node->data->key, LPID);

		// The migration message contains the state of the migrating SMH,
		//	after allocating space to locally manage the node, I've
		//	now to update the state of the SMH using the state 
		//	information contained in the migration message
		node->data->targX	= msg->migr.targX;	// X-coordinate
		node->data->targY	= msg->migr.targY;	// Y-coordinate
		node->data->Seed	= msg->migr.Seed;	// Seed for random generation
		node->data->rwp_state	= msg->migr.rwp_state;	// random way point state
	}
}
/*---------------------------------------------------------------------------*/


/* ************************************************************************ */
/* 			   	    U T I L S				    */
/* ************************************************************************ */

/*
	Simulatiom trace used by the graphic visualizer
*/
static void UNUSED PrintStep_Visualizer()
{
	int		h;
	hash_node_t	*node;


	fprintf(stdout, "<BEGIN %6.4f>\n",  simclock);

	for ( h = 0; h < table->size; h++ ) {
		for ( node = table->bucket[h]; node; node = node->next ) {
			if( node->data->changed == YES ) {

				// Printed fields:
				//	node ID
				//	LP in which the node is allocated
				//	spatial coordiantes (x, y)
				fprintf(stdout, "%d\t%d\t%10.2f\t%10.2f\n", node->data->key, node->data->lp, node->data->posX, node->data->posY);
				node->data->changed = NO;
			}
		}
	}
	
	fprintf(stdout, "<END>\n");
	fflush(stdout);
}


/*
	Loading the configuration file of the simulation
*/
static void UNUSED LoadINI(char *ini_name)
{
	int	ret;	
	char	data[64];
	

	ret = INI_Load(ini_name);
	ASSERT( ret == INI_OK, ("Error loading the .ini file \"%s\"", ini_name) );
	
	/* SIMA */
        ret = INI_Read( "SIMA", "HOST", data);
	if ( ret == INI_OK && strlen(data) )
		strcpy(SIMA_HOST, data);

        ret = INI_Read( "SIMA", "PORT", data);
	if ( ret == INI_OK && strlen(data) )
		SIMA_PORT = atoi(data);

	/* MODEL */
        ret = INI_Read( "MODEL", "SPEED", data);
	if ( ret == INI_OK && strlen(data) )
		SPEED = atoi(data);

        ret = INI_Read( "MODEL", "PAUSE", data);
	if ( ret == INI_OK && strlen(data) )
	PAUSE = atoi(data);
		
        ret = INI_Read( "MODEL", "RADIUS", data);
	if ( ret == INI_OK && strlen(data) )
		RADIUS = atof(data);

        ret = INI_Read( "MODEL", "MAX_X", data);
	if ( ret == INI_OK && strlen(data) )
		MAX_X = atof(data);

        ret = INI_Read( "MODEL", "MAX_Y", data);
	if ( ret == INI_OK && strlen(data) )
		MAX_Y = atof(data);

	INI_Free(); 
}
/*---------------------------------------------------------------------------*/


/* ************************************************************************ */
/* 			   	    M A I N				    */
/* ************************************************************************ */
int main(int argc, char* argv[])
{
	char 	msg_type, 		// Type of message
		*data,			// Buffer for incoming messages, dynamic allocation
		out[100];	 	// Output buffer

	int	count, 			// Number of SMHs to simulate in the local LP
		start,			// First identifier (ID) to be used to tag the locally managed SMHs 
		max_data;		// Maximum size of incoming messages

	int	from,			// ID of the message sender 
		to,			// ID of the message receiver 
		tot = 0;		// Total number of executed migrations

	int 	loc,			// Number of messages with local destination (intra-LP)
		rem, 			// Number of messages with remote destination (extra-LP)
		migr,			// Number of executed migrations 
		t;			// Total number of messages (local + remote)

	unsigned long	tot_loc = 0,	// Total number of messages with local destination (intra-LP), up to the current timestep
			tot_rem = 0,	// Total number of messages with remote destination (inter-LP), up to the current timestep
			tot_t = 0;	// Total number of messages (local + remote), up to the current timestep
		
	double  Ts;			// Timestep
	Msg	*msg;			// Message

	int	migrated_in_this_step;	// Number of entities migrated in this step, in the local LP

	// Time measurement
	struct timeval 	t1,t2;		

	
	// Loading the input parameters from the configuration file
	LoadINI( "wireless.ini" );

	// Returns the standard host name for the current machine
	gethostname(LP_HOST, 64);

	// Command-line input parameters
	NLP		= atoi(argv[1]);
	NSIMULATE	= atoi(argv[2]);
	
	// Proximity detection mechanism, squares optimization
	// for more details: http://doi.acm.org/10.1145/1164717.1164727
	#ifdef USE_SQUARE
	// Computes the side of the internal square
	RADIUS_INT	= (RADIUS * sqrtf(2.0));
	#endif
	
	
	/*
		Set-up of the GAIA framework

		Parameters:	
		1. (SIMULATE*NLP) 	Maximum number of simulated entities
		2. (NLP)	  	Number of LPs in the simulation
		3. (rnd_file)     	Seeds file for the random numbers generator 
		4. (NULL)         	LP canonical name
		5. (SIMA_HOST)	 	Hostname where the SImulation MAnager is running
		6. (SIMA_PORT)	  	Sima TCP port number
	*/
	LPID = GAIA_Initialize(NSIMULATE*NLP, NLP, rnd_file, NULL, SIMA_HOST, SIMA_PORT);
	
	// Returns the length of the timestep
	//	this value is defined in the "CHANNELS.TXT" configuration file
	//	given that GAIA is based on the time-stepped synchronization algorithm
	//	it retuns the size of a step
	step = GAIA_GetStep();

	// Due to synchronization constraints The FLIGHT_TIME has to be bigger than the timestep size
	if ( FLIGHT_TIME < step) {

		fprintf(stdout, "FATAL ERROR, the FLIGHT_TIME (%8.2f) is than the timestep size (%8.2f)\n", FLIGHT_TIME, step);
		fflush(stdout);
		exit(-1);
	}

	// First identifier (ID) of SMHs allocated in the local LP
	start	= NSIMULATE * LPID;

	// Number of SMHs to allocate in the local LP
	count	= NSIMULATE;

	//  Used to set the ID of the first entity (SMH) simulated by the local LP
	GAIA_SetFstID(start);

	// Used to enable the migration mechanism (GAIA)
	GAIA_SetMigration(MIGR_ON);

	// Some tuning of the migration mechanism
	// http://pads.cs.unibo.it/dokuwiki/doku.php?id=pads:gaia-apis
	GAIA_SetMF(3.0);

	// The Load Balancing mechanism is turned OFF
	GAIA_SetLoadBalancing(LOAD_OFF);

	// Redirection of standard output
	sprintf(out, "%d.out", LPID);
	fclose(stdout); stdout = fopen(out, "w"); fflush(stdout);
	
	// Data structures initialization (hash tables and migration list)
	hash_init(table, NSIMULATE * NLP);	// Hashtable of global SMHs
	hash_init(stable, NSIMULATE);		// Hashtable of local SMHs
	list_init(mlist);			// Migration list (pending migrations in the local LP)

	// Starting the execution timer
	TIMER_NOW(t1);

	fprintf(stdout, "#LP [%d] HOSTNAME [%s]\n", LPID, LP_HOST);
	fprintf(stdout, "#LP [%d] STARTED\n#\n", LPID);
	fprintf(stdout, "#LP [%d] PARAMETERS:\n", LPID);
	fprintf(stdout, "#LP [%d] \t(SPEED=%d)\n", LPID, SPEED);
	fprintf(stdout, "#LP [%d] \t(PAUSE=%d)\n", LPID, PAUSE);
	fprintf(stdout, "#LP [%d] \t(MAX_X=%.2f)\n", LPID, MAX_X);
	fprintf(stdout, "#LP [%d] \t(MAX_Y=%.2f)\n", LPID, MAX_Y);
	fprintf(stdout, "#LP [%d] \t(RADIUS=%.2f)\n#\n", LPID, RADIUS);
	fprintf(stdout, "#LP [%d]: generating SMH from %d to %d ... ", LPID, (LPID * NSIMULATE), ((LPID * NSIMULATE) + NSIMULATE) - 1);
	fflush(stdout);

	// Generate the locally managed SMHs
	Generate(count);
	fprintf(stdout, " OK\n#\n");
	
	// Only one LP is responsible to display extra information and statistics
	if ( LPID == LP_STAT ) {
		#ifdef VISUALIZER
		fprintf(stdout, "<PARAMETERS RADIUS=%.0f RING=%.0f>\n",  RADIUS, MAX_X);
		fflush(stdout);
		#endif
	}

	// Dynamically allocating some space to receive messages
	data = malloc(BUFFER_SIZE);
	ASSERT(( data != NULL ), ("simulation main: malloc error, receiving buffer NOT allocated!"));

	// The master LP has more data to print out
	if ( LPID == LP_STAT ) {
		fprintf(stdout, "#[01]: elapsed Wall-Clock-Time up to this step\n");
		fprintf(stdout, "#[02]: timestep number\n");
		fprintf(stdout, "#[03]: number of entities allocated in this LP\n");
		fprintf(stdout, "#[04]: number of entities migrated out of this LP in this timestep, local value\n");
		fprintf(stdout, "#[05]: percentage of local communications (intra-LP), global value\n");
		fprintf(stdout, "#[06]: percentage of remote communications (inter-LP), global value\n");
		fprintf(stdout, "#[07]: average percentage (up to the current timestep) of local communications (intra-LP), global value\n");
		fprintf(stdout, "#[08]: average percentage (up to the current timestep) of remote communications (intra-LP), global value\n");		
		fprintf(stdout, "#[09]: number of migrations in this timestep, global value\n");
		fprintf(stdout, "#[10]: total number of migrations up to now, global value\n");
		fprintf(stdout, "#[11]: migration ratio = average number of migrations, per entity, in 1000 timesteps\n#\n");
	} else {
		fprintf(stdout, "#[01]: elapsed Wall-Clock-Time up to this step\n");
		fprintf(stdout, "#[02]: timestep number\n");
		fprintf(stdout, "#[03]: number of entities allocated in this LP\n");
		fprintf(stdout, "#[04]: number of entities migrated out of this LP in this timestep, local value\n#\n");
	}
	
	/* Main simulation loop, receives messages and calls the handler associated with them */
	while ( !end_reached ) {
		// Max size of the next message. 
		// 	after the receive the variable will contain the real size of the message
		max_data = BUFFER_SIZE;

		// Looking for a new incoming message
		msg_type = GAIA_Receive(&from, &to,  &Ts, (void *)data, &max_data);
		msg 	 = (Msg *)data;

		// A message has been received, process it (calling appropriate handler)
		// 	message handlers
		switch (msg_type) {
			
			// The simulator topology is changing, one LP is existing from
			//	the simulation before the ending
			case STATE_CHANGE:
				// If it is the LP that manages the statistics then
				//	a new LP has to be found	
				if ( from == LP_STAT )	LP_STAT	+= 1;
				// If it is the local LP then we have to trigger some reactions
				if ( from == LPID )	to_leave = 1;
			break;

			// The migration of a locally managed SMH has to be done,
			//	calling the appropriate handler to insert the SMH identifier
			//	in the list of pending migrations
			case NOTIF_MIGR:
				notify_migration_event_handler(from, to);
			break;

			// A migration has been executed in the simulation but the local
			//	LP is not directly involved in the migration execution
			case NOTIF_MIGR_EXT:
				notify_ext_migration_event_handler(from, to);
			break;

			// Registration of a new SMH that is manager by another LP
			case REGISTER:
				register_event_handler(from, to);
			break;

			// The local LP is the receiver of a migration and therefore a new
			//	SMH has to be managed in this LP. The handler is responsible
			//	to allocate the necessary space in the LP data structures 
			//	and in the following to copy the SMH state that is contained 
			//	in the migration message
			case EXEC_MIGR:
				migration_event_handler(from, msg);
			break;

			// End Of Step:
			//	the current simulation step is finished, some pending operations
			//	have to be performed
			case EOS:
				// Stopping the execution timer 
				//	(to record the execution time of each timestep)
				TIMER_NOW(t2);

				/*  Actions to be done at the end of each simulated timestep  */
				if ( simclock < END_CLOCK ) {	// The simulation is not finished

					// The SMHs update their position in the simulated arena following
					//	a predefined mobility model, in this case Random Waypoint
					ScanActivity();

					// Simulating the wireless interactions among SMHs
					//
					//	in the last (END_CLOCK - FLIGHT_TIME) timesteps
					//	no pings will be sent because we wanna check if all
					//	sent pings are correctly received
					if ( simclock < ( END_CLOCK - FLIGHT_TIME ) )
						PingRange();

					// The pending migration of "flagged" SMHs has to be executed,
					//	the SMH to be migrated were previously inserted in the migration
					//	list due to the receiving of a "NOTIF_MIGR" message sent by 
					//	the GAIA framework
					migrated_in_this_step = ScanMigrating();

					// If VISUALIZER is defined then print the position of each node at every step
					#ifdef VISUALIZER				
					if ( LPID == LP_STAT )
						PrintStep_Visualizer();
					#endif

					// The LP that manages statistics prints out them
					if ( LPID == LP_STAT ) {		// Verbose output	

						// Some of them are provided by the GAIA framework
						GAIA_GetStatistics(&loc, &rem, &migr);

						// Total number of interactions (in the timestep)
						t 	= loc + rem;

						// Computing some totals (up to this timestep)
						tot_loc += loc;		// local interations
						tot_rem += rem;		// remote interations
						tot_t += t;		// local + remote interactions
						
						// Total number of migrations (in the simulation run)
						tot	+= migr;

						// Printed fields:
						// 	elapsed Wall-Clock-Time up to this step
						//	timestep number
						//	number of entities in this LP
						//	number of entities migrated out of this LP in this timestep, local value
						//	percentage of local communications (intra-LP), global value
						//	percentage of remote communications (inter-LP), global value
						//	average percentage (up to the current timestep) of local communications (intra-LP), global value
						//	average percentage (up to the current timestep) of remote communications (intra-LP), global value
						//	number of migrations in this timestep, global value
						//	total number of migrations up to now, global value
						//	migration ratio = average number of migrations, per entity, in 1000 timesteps
						fprintf(stdout, "[%11.2f]\t[%6.5f]\t%4.0f\t%4d\t%2.2f\t%2.2f\t%2.2f\t%2.2f\t%d\t%d\t%2.2f\n", TIMER_DIFF(t2,t1), simclock, (float)stable->count, migrated_in_this_step, (float)loc / (float)t * 100.0, (float)rem / (float)t * 100.0, (float)tot_loc / (float)tot_t * 100.0, (float)tot_rem / (float)tot_t * 100.0, migr, tot, (float)tot / (NSIMULATE * NLP * ( simclock / 1000)));

						fflush(stdout);
					}
					else {	// Reduced output
	
						fprintf(stdout, "[%11.2fs]   %12.2f [%d]\t[%4d]\n", TIMER_DIFF(t2,t1), simclock, stable->count, migrated_in_this_step);
						fflush(stdout);
					}

					// Now it is possible to advance to the next timestep
					simclock = GAIA_TimeAdvance();

					if ( to_leave ) {
						/* Premature exit from the simulation (due to performance reasons) */
						TIMER_NOW(t2);
		
						fprintf(stdout, "\n\n");
						fprintf(stdout, "#-- Termination condition reached\n");
						fprintf(stdout, "#-- Elapsed time (Wall Clock Time): %.2f\n",TIMER_DIFF(t2,t1));
						fprintf(stdout, "#-- Total sent pings (by this LP): %ld\n", lp_total_sent_pings);
						fprintf(stdout, "#-- Total received pings (by this LP): %ld\n", lp_total_received_pings);
						fflush(stdout);	

						end_reached = 1;
					}
				}
				else {
					/* End of simulation */
					TIMER_NOW(t2);
	
					fprintf(stdout, "\n\n");
					fprintf(stdout, "#-- Termination condition reached. Clock: %.2f\n", simclock);

					// The LP that manages statistics prints out them
					if ( LPID == LP_STAT ) {		// Verbose output	
						fprintf(stdout, "#-- Total number of migrations (global value): %d\n", tot);		
						fprintf(stdout, "#-- Final average local communication ratio: %2.2f\n", (float)tot_loc / (float)tot_t * 100.0);
						fprintf(stdout, "#-- Final average local communication ratio gain: %2.2f\n", ((float)tot_loc / (float)tot_t * 100.0) - (float)(100.0 / NLP));
						fprintf(stdout, "#-- Final average remote communication ratio: %2.2f\n", (float)tot_rem / (float)tot_t * 100.0);
						fprintf(stdout, "#-- Final migration ratio: %2.2f\n", (float)tot / (NSIMULATE * NLP * ( simclock / 1000)));	
					}
					
					fprintf(stdout, "#-- Elapsed time (Wall Clock Time): %.2f\n",TIMER_DIFF(t2,t1));
					fprintf(stdout, "#-- Total sent pings (by this LP): %ld\n", lp_total_sent_pings);
					fprintf(stdout, "#-- Total received pings (by this LP): %ld\n", lp_total_received_pings);
					fflush(stdout);	
	
					end_reached = 1;
				}
			break;

			// Simulated model events (user level events)
			case UNSET:		
				// Updates the SMH position
				if ( msg->type == 'M' )	{
					move_event_handler(from, msg);
				}
				
				// Wireless device turned ON of OFF
				if ( msg->type == 'W' )	{
					comm_event_handler(from, msg);
				}

				// Ping message
				if ( msg->type == 'P' )	{
					ping_event_handler(from, to, msg);
				}
			break;

			default:
				fprintf(stdout, "Unknown event\n");
		}
	}

	// Finalize the GAIA framework
	GAIA_Finalize();
	
	// Freeing of the receiving buffer
	free(data);

	return 0;
}
